const char kPathSeparator =
#ifdef _WIN32
'\\';
#else
'/';
#endif

#include <fstream>
#include <sstream>
#include <iomanip>
#include <cstdlib>
#include <cstring>
#include <cuda.h>

#include "cuda_check.cuh"
#include "dim.cuh"
#include "user.cuh"

#include "values.hxx"
#include "os.hxx"
#include "timer.hxx"

__global__ void dprint(double *u, int n0, int n1, int n2)
{
  int i, j, k;
  int p;

  for (k=0; k<n2; k++) {
    for (j=0; j<n1; j++) {
      for (i=0; i<n0; i++) {
        p = IDX(i,j,k);
        printf(" %f", u[p]);
      }
      printf("\n");
    }
    printf("\n");
  }
}

__global__
void zeroValue(double *u, int n)
{
  int i = threadIdx.x + blockIdx.x * blockDim.x;
  if (i<n) {
    u[i] = 0.0;
  }
}

__global__
void initValue(double *u, int n0, int n1, int n2)
{
  int i = threadIdx.x + blockIdx.x * blockDim.x;
  int j = threadIdx.y + blockIdx.y * blockDim.y;
  int k = threadIdx.z + blockIdx.z * blockDim.z;
  int p;

  if (i>0 && i<n0-1 && j>0 && j<n1-1 && k>0 && k<n2-1)
  {
    p = IDX(i,j,k);
    u[p] = cond_ini(d_xmin[0] + i*d_dx[0],
                    d_xmin[1] + j*d_dx[1], 
                    d_xmin[2] + k*d_dx[2]);
  }
}

__global__
void boundZValue(int k, double *u, int n0, int n1, int n2)
{
  int i = threadIdx.x + blockIdx.x * blockDim.x;
  int j = threadIdx.y + blockIdx.y * blockDim.y;
  int p;

  if (i<n0 && j<n1) {
    p = IDX(i,j,k);
    u[p] = cond_ini(d_xmin[0] + i*d_dx[0], 
                    d_xmin[1] + j*d_dx[1], 
                    d_xmin[2] + k*d_dx[2]);    
  }
}


__global__
void boundYValue(int j, double *u, int n0, int n1, int n2)
{
  int i = threadIdx.x + blockIdx.x * blockDim.x;
  int k = threadIdx.z + blockIdx.z * blockDim.z;
  int p;

  if (i<n0 && k<n1) {
    p = IDX(i,j,k);
    u[p] = cond_ini(d_xmin[0] + i*d_dx[0], 
                    d_xmin[1] + j*d_dx[1], 
                    d_xmin[2] + k*d_dx[2]);
  }
}

__global__
void boundXValue(int i, double *u, int n0, int n1, int n2)
{
  int j = threadIdx.y + blockIdx.y * blockDim.y;
  int k = threadIdx.z + blockIdx.z * blockDim.z;
  int p;

  if (j<n1 && k<n2) {
    p = IDX(i,j,k);
    u[p] = cond_ini(d_xmin[0] + i*d_dx[0], 
                    d_xmin[1] + j*d_dx[1], 
                    d_xmin[2] + k*d_dx[2]);
  }
}

Values::Values(Parameters & prm) : m_p(prm)
{
  int i;

  for (i=0; i<3; i++) {
    m_imin[i] = m_p.imin(i);
    m_imax[i] = m_p.imax(i);
    m_n[i] = m_imax[i] - m_imin[i] + 3;
    m_dx[i] = m_p.dx(i);
    m_xmin[i] = m_p.xmin(i);
    m_xmax[i] = m_p.xmax(i);
  }

  n0 = m_n[0];
  n1 = m_n[1];
  n2 = m_n[2];
  n_total = n0 * n1 * n2;
  n_bytes = n_total * sizeof(double);
  
  // A completer : reservation de la memoire de la classe value
  // - m_hu : tableau de doubles de taille n0 * n1 * n2 dans la memoire du CPU
  // - m_du : tableau de doubles de taille n0 * n1 * n2 dans la memoire du GPU

  h_synchronized = false;
}


Values::~Values()
{
// A compléter :: libération de la mémoire de la classe value 
}

void Values::zero()
{
  int dimBlock = 256;
  int dimGrid = ceil(n_total/double(dimBlock));

  zeroValue<<<dimGrid, dimBlock>>>(m_du, n_total);
  CUDA_CHECK_KERNEL();

  h_synchronized = false;
}

void Values::init()
{
  dim3 dimBlock(8,8,8);
  dim3 dimGrid(int(ceil(n0/double(dimBlock.x))),
               int(ceil(n1/double(dimBlock.y))),
               int(ceil(n2/double(dimBlock.z))));

  initValue<<<dimGrid, dimBlock>>>(m_du, n0, n1, n2);
  CUDA_CHECK_KERNEL();

  h_synchronized = false;
}

void Values::boundaries()
{
  dim3 dimBlock1(16,16,1);
  dim3 dimGrid1(int(ceil(n0/double(dimBlock1.x))),
                int(ceil(n1/double(dimBlock1.y))), 
                1);
  boundZValue<<<dimGrid1, dimBlock1>>>(m_imin[2]-1, m_du, n0, n1, n2);
  boundZValue<<<dimGrid1, dimBlock1>>>(m_imax[2]+1, m_du, n0, n1, n2);

  dim3 dimBlock2(16,1,16);
  dim3 dimGrid2(int(ceil(n0/double(dimBlock2.x))), 
                1,
                int(ceil(n2/double(dimBlock2.z))));

  boundYValue<<<dimGrid2, dimBlock2>>>(m_imin[1]-1, m_du, n0, n1, n2);
  boundYValue<<<dimGrid2, dimBlock2>>>(m_imax[1]+1, m_du, n0, n1, n2);

  dim3 dimBlock3(1,16,16);
  dim3 dimGrid3(1, 
                int(ceil(n1/double(dimBlock3.y))),
                int(ceil(n2/double(dimBlock3.z))));

  boundXValue<<<dimGrid3, dimBlock3>>>(m_imin[0]-1, m_du, n0, n1, n2);
  boundXValue<<<dimGrid3, dimBlock3>>>(m_imax[0]+1, m_du, n0, n1, n2);

  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();

  h_synchronized = false;
}

std::ostream & operator<< (std::ostream & f, Values & v)
{
  v.print(f);
  return f;
}

void Values::print(std::ostream & f)
{
  copyGPUtoCPU();

  int i, j, k;
  int imin = m_imin[0]-1, jmin = m_imin[1]-1, kmin = m_imin[2]-1;
  int imax = m_imax[0]+1, jmax = m_imax[1]+1, kmax = m_imax[2]+1;

  for (k=kmin; k<=kmax; k++) {
    for (j=jmin; j<=jmax; j++) {
      for (i=imin; i<=imax; i++) {
        f << " " << std::setw(12) << std::setprecision(6) << operator()(i,j,k);
      }
      f << std::endl;
    }
    f << std::endl;
  }
}

template<typename T>
void swap(T & a, T & b)
{
  T t = a;
  a = b;
  b = t;
}

void Values::swap(Values & other)
{
  ::swap(m_du, other.m_du);
  ::swap(m_hu, other.m_hu);
  ::swap(h_synchronized, other.h_synchronized);
  int i;
  for (i=0; i<3; i++) {
    ::swap(m_imin[i], other.m_imin[i]);
    ::swap(m_imax[i], other.m_imax[i]);
    ::swap(m_n[i], other.m_n[i]);
    ::swap(m_dx[i], other.m_dx[i]);
    ::swap(m_xmin[i], other.m_xmin[i]);
    ::swap(m_xmax[i], other.m_xmax[i]);
  }
  ::swap(n0, other.n0);
  ::swap(n1, other.n1);
  ::swap(n2, other.n2);
  ::swap(n_total, other.n_total);
  ::swap(n_bytes, other.n_bytes);
  
}

void Values::plot(int order) {

  copyGPUtoCPU();

  std::ostringstream s;
  int i, j, k;
  int imin = m_imin[0]-1, jmin = m_imin[1]-1, kmin = m_imin[2]-1;
  int imax = m_imax[0]+1, jmax = m_imax[1]+1, kmax = m_imax[2]+1;

  s << m_p.resultPath();
  mkdir_p(s.str().c_str());

  s << kPathSeparator << "plot_" << std::setw(5) << std::setfill('0') << order << ".vtr";
  std::ofstream f(s.str().c_str());

  f << "<?xml version=\"1.0\"?>\n";
  f << "<VTKFile type=\"RectilinearGrid\" version=\"0.1\" byte_order=\"LittleEndian\">\n"
    << "<RectilinearGrid WholeExtent=\""
    << imin << " " << imax  << " "
    << jmin << " " << jmax  << " "
    << kmin << " " << kmax
    << "\">\n"
    << "<Piece Extent=\""
    << imin << " " << imax  << " "
    << jmin << " " << jmax  << " "
    << kmin << " " << kmax
    << "\">\n";

  f << "<PointData Scalars=\"values\">\n";
  f << "  <DataArray type=\"Float64\" Name=\"values\" format=\"ascii\">\n";
  
  for (k=kmin; k<=kmax; k++) {
    for (j=jmin; j<=jmax; j++) {
      for (i=imin; i<=imax; i++)
        f << " " << std::setw(7) << operator()(i,j,k);
      f << "\n";
    }
    f << "\n";
  }
  f << " </DataArray>\n";
   
  f << "</PointData>\n";

  f << " <Coordinates>\n";
  
  for (k=0; k<3; k++) {
    f << "   <DataArray type=\"Float64\" Name=\"" << char('X' + k) << "\"" 
      << " format=\"ascii\">";
    
    int imin = m_p.imin(k)-1;
    int imax = m_p.imax(k)+1;
    for (i=imin; i<=imax; i++)
      f << " " << i * m_p.dx(k);
    f << "   </DataArray>\n";
  }
  f << " </Coordinates>\n";
  
  f << "</Piece>\n"
    << "</RectilinearGrid>\n"
    << "</VTKFile>\n" <<std::endl;
}

void Values::operator= (const Values &other)
{
  int i;
  
  n0 = other.n0;
  n1 = other.n1;
  n2 = other.n2;
  n_total = other.n_total;
  n_bytes = other.n_bytes;
  for (i=0; i<3; i++) {
    m_imin[i] = other.m_imin[i];
    m_imax[i] = other.m_imax[i];
    m_n[i] = other.m_n[i];
    m_xmin[i] = other.m_xmin[i];
    m_xmax[i] = other.m_xmax[i];
    m_dx[i] = other.m_dx[i];
  }
  h_synchronized = other.h_synchronized;

  if (other.h_synchronized)
    memcpy(m_hu, other.m_hu, n_bytes);
  cudaMemcpy(m_du, other.m_du, n_bytes, cudaMemcpyDeviceToDevice);
}

void Values::copyGPUtoCPU()
{
  if (!h_synchronized) {
    cudaMemcpy(m_hu, m_du, n_bytes, cudaMemcpyDeviceToHost);
    h_synchronized = true;
  }

}