
#include "cuda_check.cuh"
#include "variation.hxx"

#include <stdio.h>
#include <iostream>
#include <vector>
#include <cooperative_groups.h>

namespace cg = cooperative_groups;

#include "cuda_check.cuh"
#include "timer.hxx"
#include "values.hxx"

__global__ void
difference (const double *u, 
            const double *v, 
            double * duv, int n)
{
  int i = threadIdx.x + blockIdx.x * blockDim.x;
  if (i<n) {
    duv[i] = fabs(u[i] - v[i]);
  }   
}

__global__ void
reduce(const double *u, double *partialDiff, int n)
{
  cg::thread_block cta = cg::this_thread_block();
  extern __shared__ double sdata[];
  
  int tid = threadIdx.x;
  int bid = blockIdx.x;
  int i = bid*blockDim.x + tid;

  sdata[tid] = (i<n) ? u[i] : 0.0 ;
  cg::sync(cta);

  for (unsigned int s=blockDim.x/2; s>0; s>>=1)
    {
      if (tid < s)
        sdata[tid] += sdata[tid + s];

      cg::sync(cta);
    }
  cg::sync(cta);
  if (tid == 0) {
    partialDiff[blockIdx.x] = sdata[0];
  }
}

__global__ void printU(const double *u, int n)
{
  int i;
  for (i=0; i<n; i++)
    printf("u[%d] = %g\n", i, u[i]);
}

double variation(const Values &u, 
                        const Values &v,
                        Values & diff)
{
  int dimBlock = 64;
  int dimGrid = int(ceil(u.size()/double(dimBlock)));
  
  int nbytesBlock = dimBlock * sizeof(double);

  double * pdiff = diff.device_data();

  double * pdiffPartial;
  CUDA_CHECK_OP(cudaMalloc(&pdiffPartial, dimGrid*sizeof(double)));
  
  difference<<< dimGrid, dimBlock>>> 
     (u.device_data(), v.device_data(), pdiff, u.size()); 
  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();

  reduce<<< dimGrid, dimBlock, nbytesBlock >>>
     (pdiff, pdiffPartial, diff.size());
  cudaDeviceSynchronize();
  CUDA_CHECK_KERNEL();

  std::vector<double> h_partialSums(dimGrid);
  cudaMemcpy(h_partialSums.data(), pdiffPartial, dimGrid* sizeof(double), cudaMemcpyDeviceToHost);

  double s = 0.0;
  for (int i=0; i<dimGrid; i++)
    s += h_partialSums[i];

  cudaFree(pdiffPartial);
  return s;
}

